from typing import Any, Callable, List, Union, Tuple
from pathlib import Path
import pandas as pd
import pyarrow.parquet as pq


class Read:
    def __init__(self, fn: Callable, *args, **kwds) -> None:
        self._fn = fn
        self._args = args
        self._kwds = kwds

    def __call__(self, path: Any, *args, **kwds) -> Any:
        return self._fn(path, *self._args, *args, **self._kwds, **kwds)


class FileRead(Read):
    def __init__(self, fn: Callable, add_path: bool = False, *args, **kwds) -> None:
        super().__init__(fn, *args, **kwds)
        self._add_path = add_path

    def __call__(self, path: Path, *args, **kwds) -> Union[Any, Tuple[Any, Path]]:
        result = super().__call__(path, *args, **kwds)
        if self._add_path:
            return result, path
        return result


class ParquetRead(FileRead):
    def __init__(self, *args, **kwds) -> None:
        super().__init__(self._fn, *args, **kwds)

    def _fn(self, path: Path, *args, **kwds) -> pd.DataFrame:
        return pd.read_parquet(path, *args, **kwds)


class ParquetMetaDataRead(FileRead):
    def __init__(self, *args, **kwds) -> None:
        super().__init__(self._fn, *args, **kwds)

    def _fn(self, path: Path, *args, **kwds) -> pq.FileMetaData:
        return pq.read_metadata(path, *args, **kwds)


class MultiRead:
    def __init__(self, read: Read, *args, **kwds) -> None:
        self._read = read
        self._args = args
        self._kwds = kwds

    def __call__(self, paths: List[Path], *args, **kwds) -> List[Any]:
        return [self._read(
            path, *self._args, *args, **self._kwds, **kwds
            ) for path in paths]


def find_paths(
    parent: Union[str, Path],
    pattern: str = '*',
    recursive: bool = False
    ) -> Union[List[Path], List[List[Path]]]:

    if isinstance(parent, str):
        parent = Path(parent)
    if recursive:
        paths = [p for p in parent.rglob(pattern) if p.is_file()]
        result = []
        for parent in {p.parent for p in paths}:
            result.append([p for p in paths if p.parent == parent])
        return result
    else:
        return [p for p in parent.glob(pattern) if p.is_file()]


import pytest
from pathlib import Path
import pandas as pd


@pytest.fixture(scope='session')
def fib_df():
     return pd.DataFrame({
        'time': [
            pd.to_datetime('2022-09-09 00:00:00'),
            pd.to_datetime('2022-09-09 00:00:01'),
            pd.to_datetime('2022-09-09 00:00:01'),
            pd.to_datetime('2022-09-09 00:00:02'),
            pd.to_datetime('2022-09-09 00:00:03'),
            pd.to_datetime('2022-09-09 00:00:05'),
        ],
        'x': [0, 1, 1, 2, 3, 5]
    })


@pytest.fixture(scope='session')
def data_dir(tmp_path_factory):
    return tmp_path_factory.mktemp('data')


@pytest.fixture(scope='session')
def sample_paths(data_dir, fib_df):
    paths = [data_dir / Path(path) for path in [
        'file_1.parquet',
        'file_2.parquet'
    ]]
    for path in paths:
        fib_df.to_parquet(path)
    return paths


@pytest.fixture(scope='session')
def sample_path(sample_paths):
    return sample_paths[0]



import numpy as np

from oocas.cache import Cache, IndexCache, TimeIndexCache


def test_cache(fib_df):
    fn = Cache()
    fn(fib_df)
    assert fn.data.equals(fib_df)


def test_indexcache(fib_df):
    fn = IndexCache(1)
    fn(fib_df)
    assert np.array_equal(
        fn.data['x'].to_numpy(), np.array([5]))


def test_timeindexcache(fib_df):
    fn = TimeIndexCache('4S')
    fn(fib_df.set_index('time'))
    assert np.array_equal(
        fn.data['x'].to_numpy(), np.array([1, 1, 2, 3, 5]))


from oocas.read import Read, FileRead, ParquetRead, ParquetMetaDataRead, MultiRead, find_paths


def test_read():
    fn = Read(
        lambda x, suffix: x + suffix,
        '.csv'
    )
    assert fn('file1') == 'file1.csv'


def test_fileread():
    fn = FileRead(
        lambda x, suffix: x + suffix,
        add_path = True,
        suffix='.csv'
    )
    assert fn('file1') == ('file1.csv', 'file1')


def test_parquetread(sample_path, fib_df):
    fn = ParquetRead()
    df = fn(sample_path)
    assert df.equals(fib_df)


def test_parquetmetadataread(sample_path):
    fn = ParquetMetaDataRead()
    assert fn(sample_path).num_rows == 6


def test_multiread():
    fn = MultiRead(
        FileRead(
            lambda x, suffix: x + suffix,
            add_path=True),
    )
    assert fn(('file1', 'file2'), '.csv')\
        == [('file1.csv', 'file1'), ('file2.csv', 'file2')]


def test_findpaths(data_dir, sample_paths):
    paths = find_paths(data_dir, recursive=True)
    assert paths[0] == sample_paths


    


from typing import Any, Callable, Union
import pandas as pd


class Cache:
    def __init__(self,
        fn: Callable = lambda x: x,
        data: Any = None,
        *args, **kwds) -> None:

        self._fn = fn
        self.data = data
        self._args = args
        self._kwds = kwds

    def __call__(self, data: Any, *args, **kwds) -> None:
        self.data = self._fn(data, *self._args, *args, **self._kwds, **kwds)


class IndexCache(Cache):
    def __init__(self, lookback: int) -> None:
        super().__init__(
            self._fn
        )
        self.lookback = lookback

    def _fn(self, data: pd.DataFrame, lookback: int = None) -> pd.DataFrame:
        if not lookback:
            lookback = self.lookback
        df = pd.concat([self.data, data], axis=0)
        cutoff = len(df) - lookback
        return df[cutoff:].copy()


class TimeIndexCache(Cache):
    def __init__(self, lookback: Union[str, pd.Timedelta, int]) -> None:
        super().__init__(
            self._fn
        )
        self.lookback = pd.Timedelta(lookback)

    def _fn(self, data: pd.DataFrame, lookback: pd.Timedelta = None) -> pd.DataFrame:
        if not lookback:
            lookback = self.lookback
        df = pd.concat([self.data, data], axis=0)
        cutoff = df.index[-1] - lookback
        return df[cutoff:].copy()


from oocas.transform import Transform, CacheTransform


def test_transform():
    transform = Transform(
        lambda x, suffix: x + suffix,
        '.csv'
    )
    assert transform('file1') == 'file1.csv'


def test_cachetransform():
    transform = CacheTransform(
        lambda a, cache, b: a + cache.data + b,
        2
    )
    assert transform(1, 3) == 6



from pathlib import Path

from oocas.write import Write, FileWrite, MultiWrite


def test_write():
    fn = Write(
        lambda x, suffix: x + suffix,
    )
    assert fn('file1', '.csv') == 'file1.csv'


def test_filewrite():
    fn = FileWrite(
        lambda a, _, b: a + b,
        path_transform=lambda x: x[0],
        path_part_replace=('C', 'D'),
        name_transform=lambda x: x[:-1],
        suffix='.csv',
        add_result=True,
        b=2
    )
    assert fn(1, [Path('/C/data/file1.parquet')]) == (Path('/D/data/file.csv'), 3)


def test_multiwrite():
    fn = MultiWrite(
        Write(lambda x, path: path),
    )
    assert fn((1, 2), ('file1.csv', 'file2.csv'))\
        == ['file1.csv', 'file2.csv']


from typing import Callable, List, Any
from tqdm.auto import tqdm
from datetime import datetime


class Process:
    def __init__(self,
        read: Callable = lambda x: x,
        transform: Callable = lambda x: x,
        write: Callable = lambda x, _: x,
        progress_bar=True,
        verbose: bool = False
        ) -> None:

        self._read = read
        self._transform = transform
        self._write = write
        self._verbose = verbose
        self._tqdm = tqdm if progress_bar else lambda x: x

    def __call__(self, data) -> List[Any]:
        return self._fn(data)

    def _fn(self, paths) -> List[Any]:
        results = []
        for path in self._tqdm(paths):
            if self._verbose:
                print(f"{datetime.now().strftime('%H:%M:%S')}   {path}")
            data = self._read(path)
            data = self._transform(data)
            results.append(self._write(data, path))
        return results


from typing import Any, Callable, Tuple, Union, List
import pandas as pd
from pathlib import Path


class Write:
    def __init__(self, fn: Callable, *args, **kwds) -> None:
        self._fn = fn
        self._args = args
        self._kwds = kwds

    def __call__(self, data: Any, path: Any, *args, **kwds) -> Any:
        return self._fn(data, path, *self._args, *args, **self._kwds, **kwds)


class FileWrite(Write):
    def __init__(self,
        fn: Callable = lambda _, path: path,
        path_transform: Union[Callable[[Path], Path], None] = None,
        path_part_replace: Union[Tuple[str, str], None] = None,
        name_transform: Union[Callable[[str], str], None] = None,
        suffix: Union[str, None] = None,
        mkdirs: bool = False,
        overwrite: bool = False,
        add_result: bool = False,
        *args,
        **kwds) -> None:

        self._path_transform = path_transform
        self._path_part_replace = path_part_replace
        self._name_transform = name_transform
        self._suffix = suffix
        self._mkdirs = mkdirs
        self._overwrite = overwrite
        self._add_result = add_result
        super().__init__(fn, *args, **kwds)

    def __call__(self, data: Any, path: Path, *args, **kwds) -> Path:
        if self._path_transform:
            path = self._path_transform(path)
        if self._path_part_replace:
            path = replace_path_part(path, *self._path_part_replace)
        if self._name_transform:
            path = path.with_stem(self._name_transform(path.stem))
        if self._suffix:
            path = path.with_suffix(self._suffix)
        if not self._overwrite:
            check_exists(path)
        if self._mkdirs:
            mkdirs(path)
        result = super().__call__(data, path, *args, **kwds)
        if self._add_result:
            return path, result
        return path


class ParquetWrite(FileWrite):
    def __init__(self, suffix='.parquet', *args, **kwds) -> None:
        super().__init__(self._fn, suffix=suffix, *args, **kwds)

    def _fn(self, data: pd.DataFrame, path: Path, *args, **kwds) -> None:
        data.to_parquet(path, *args, **kwds)


class MultiWrite:
    def __init__(self, write: Write, *args, **kwds) -> List[Any]:
        self._write = write
        self._args = args
        self._kwds = kwds
    
    def __call__(self, data: List[Any], paths: List[Path], *args, **kwds) -> Any:
        return [self._write(
            x, path, *self._args, *args, *self._kwds, **kwds
            ) for x, path in zip(data, paths)]


def mkdirs(path: Path) -> None:
        dir_path = Path(*path.parts[:-1])
        if not dir_path.exists():
            dir_path.mkdir(parents=True, exist_ok=True)


def check_exists(path: Path) -> None:
    if path.exists():
        raise FileExistsError(
            f"File {path} already exists")


def replace_path_part(path: Path, from_part: str, to_part: str) -> Path:
    idx = path.parts.index(from_part)
    path = Path(
        *path.parts[:idx],
        to_part,
        *path.parts[idx + 1:])
    return path


from typing import Any, Callable, Union
import oocas as oc


class Transform:
    def __init__(self, fn: Callable, *args, **kwds) -> None:
        self._fn = fn
        self._args = args
        self._kwds = kwds

    def __call__(self, data: Any, *args, **kwds) -> Any:
        return self._fn(data, *self._args, *args, **self._kwds, **kwds)


class CacheTransform(Transform):
    def __init__(self, fn: Callable, cache: Union[oc.Cache, Any] = oc.Cache(), *args, **kwds) -> None:
        super().__init__(fn, *args, **kwds)
        if not isinstance(cache, oc.Cache):
            self._cache = oc.Cache(data=cache)
        else:
            self._cache = cache

    def __call__(self, data: Any, *args, **kwds) -> Any:
        return super().__call__(data, self._cache, *args, **kwds)
